function [PAR, BRDF_r] = fitBRDFstencils(matName, naxes, scale, blend, w, k, plot1)

% J. Filip, UTIA AV CR v.v.i, filipj@utia.cas.cz
% v1.0, 20/6/2015
% fitting BRDF stencils to UTIA BRDF ---------------------------
% 
% INPUTS: matName - material name of BRDF file
%         naxes - number of anisotropic axes to fit
%         scale - scale factor of BRDFs
%         blend - vector of blending coefficient of size [1 x naxes-1]
%         w - vector of anisotropy support masks of size [1 x naxes]
%         k - multiplication coefficient (default = 1)
%         AHLshift - vectors [x,y] of highlight map shift for each anis.
%         mode default [0,0]
%         plot - toggle plotting of the figures 0/1 (default 1)
% OUTPUT: PAR - estimated parameters
%         BRDF_f - estimated BRDF
%
% Licence: Creative Commons Attribution-NonCommercial 4.0 International (CC BY-NC 4.0) 
%          https://creativecommons.org/licenses/by-nc/4.0/
%

if(~exist('scale','var') || isempty(scale))
    scale = 128.0;
end

if(~exist('blend','var') || isempty(blend))
    blend = 0.0*ones(naxes-1,1);
end;

if(~exist('w','var') || isempty(w))
    w = 0.8*ones(naxes,1);
end;

if(~exist('k','var') || isempty(k))
    k = 1.0;
end;

if(~exist('plot1','var') || isempty(plot1))
    plot1 = 1;
end

AHLshift{1} = [0,0];
AHLshift{2} = [0,0];
AHLshift{3} = [0,0];

% loading BRDF from UTIA BRDF database (http://btf.utia.cas.cz) -------------
phiStep = 7.5;
thetaStep = 15;
np = 360/phiStep;  % 48
nt = 90/thetaStep; %  6
iPath = sprintf('brdfs/%s.mat',matName);
load(iPath,'BRDF');

wc = [0.2126, 0.7152, 0.0722];
mBRDF = wc(1)*BRDF(:,:,1)+wc(2)*BRDF(:,:,2)+wc(3)*BRDF(:,:,3); % mean BRDF

tic;
% analyse intensity =============================================

if(1)
    % detect anisotropy highlights azimuths (\phi_a) ------------------------
    surr = 8;
    ss = mBRDF(end,5*np+1:end); % at the highest elevations
    for i=1:naxes
        [i1,k1] = max(ss);
        a = k1-surr;
        b = k1+surr;
        if(b>np) b = b-np+1; end;
        if(a<=0) a = a+np; end;
        if(a>b)
            ss(round(a):end) = 0;
            ss(1:round(b)) = 0;
        else
            ss(round(a):round(b)) = 0;
        end;
        
        P.th{i}.apos = k1;
        fprintf('Thread direction: %3.1f, intensity %3.1f\n',P.th{i}.apos,i1);
    end;
    [ib,kb] = min(ss);
    P.bcg.pos = kb;
else    
    % manual overriding estimated anisotropy azimuthal angles indices
    P.th{1}.apos = 1;
    P.th{2}.apos = 24;
    P.th{3}.apos = 11;
end;

% getting intensity values from BRDF -----------------------------------
% threads highlights locations (for intensity picking)
npi_th = cell(naxes);
npv_th = cell(naxes);
t_sh = cell(naxes);
for i=1:naxes
    t_sh{i} = round(AHLshift{i}/phiStep);
    npi_th{i} = round(P.th{i}.apos/2) + t_sh{i}(1);
    npv_th{i} = round(P.th{i}.apos/2) + t_sh{i}(2);
end;
% specular highlight locations
npi_spe = 1 + 1;
npv_spe = round(P.th{1}.apos/2)+np/2 + 1 + t_sh{1}(2); % the most intensive highlight

% elevation dependent intensity of threads
Ith = cell(naxes);
for i=1:naxes
    Ith{i} = zeros(nt,nt,3); 
end
Ibcg = zeros(nt,nt,3); % elevation dependent intensity of background
Ispe = zeros(nt,nt,3); % elevation dependent intensity of specular highlights


for ti=0:nt-1
    for tv=ti:nt-1
        % get threads color
        for i=1:naxes
            ni = ti*np+npi_th{i}+1;
            nv = tv*np+npv_th{i}+1;
            [u,v] = getMax(BRDF,ni,nv,np);
            Ith{i}(ti+1,tv+1,:) = BRDF(u,v,:);
            Ith{i}(tv+1,ti+1,:) = BRDF(u,v,:);
        end;
        % get specular color
        ni = ti*np+npi_spe+1;
        nv = tv*np+npv_spe+1;
        [u,v] = getMax(BRDF,ni,nv,np);
        Ispe(ti+1,tv+1,:) = BRDF(u,v,:);
        Ispe(tv+1,ti+1,:) = BRDF(u,v,:);
        % get background color
        minBRDF = min(BRDF,3);
        [~,i] = min(reshape(minBRDF(ti*np+1:(ti+1)*np,tv*np+1:(tv+1)*np),np*np,1));
        [pi,pv] = ind2sub(np,i);
        Ibcg(ti+1,tv+1,:) = BRDF(ti*np+pi,tv*np+pv,:);
        Ibcg(tv+1,ti+1,:) = BRDF(ti*np+pi,tv*np+pv,:);       
    end
end

Ibcg = min(Ibcg,Ith{2});

mIth = cell(naxes);
thIntensity = zeros(naxes,1);
for i=1:naxes
    if(0)
        mIth{i} = wc(1)*Ith{i}(:,:,1)+wc(2)*Ith{i}(:,:,2)+wc(3)*Ith{i}(:,:,3);
    else
        [~,ind] = sort([Ith{i}(nt,nt,1),Ith{i}(nt,nt,2),Ith{i}(nt,nt,3)],'descend');
        mIth{i} = Ith{i}(:,:,ind(1));
    end;
    thIntensity(i) = mean(reshape(mIth{i},nt*nt,1));
end;
mIbcg = wc(1)*Ibcg(:,:,1) + wc(2)*Ibcg(:,:,2) + wc(3)*Ibcg(:,:,3);
mIspe = wc(1)*Ispe(:,:,1) + wc(2)*Ispe(:,:,2) + wc(3)*Ispe(:,:,3);

% compute thread order, th1 - the top one (the highest mean intensity)
[~,thOrder] = sort(thIntensity,'descend');

km = reshape(mIth{thOrder(1)},6*6,1)\reshape(mIspe,6*6,1);
P.spe_th1.r = [0,km];

for i=1:naxes
    thCol = scale*[Ith{thOrder(i)}(nt,nt,1),Ith{thOrder(i)}(nt,nt,2),Ith{thOrder(i)}(nt,nt,3)];
    fprintf('Thread %d color: %f %f %f\n',i,thCol(1),thCol(2),thCol(3));
    P.th{thOrder(i)}.col = thCol;
end;
bcgCol = scale*[median(median(Ibcg(4:nt,4:nt,1))),median(median(Ibcg(4:nt,4:nt,2))),median(median(Ibcg(4:nt,4:nt,3)))];
fprintf('Background color: %f %f %f\n',bcgCol(1),bcgCol(2),bcgCol(3));
speCol = scale*[mean2(Ispe(4:nt,4:nt,1)),mean2(Ispe(4:nt,4:nt,2)),mean2(Ispe(4:nt,4:nt,3))];
fprintf('Specular color: %f %f %f\n',bcgCol(1),bcgCol(2),bcgCol(3));
fprintf('Specular vs. th1 ratio: %f %f\n',P.spe_th1.r(1),P.spe_th1.r(2));

P.bcg.col = bcgCol;
P.spe.col = speCol;
P.naxes = naxes;

% fitting cosine to intensity data of each thread ------------
[ti,tv] = meshgrid(0:thetaStep:thetaStep*(nt-1));
x = reshape(ti,nt*nt,1);
y = reshape(tv,nt*nt,1);
elLim = 30;
f = cell(naxes);
for j=1:naxes
    i = thOrder(j);
    z = reshape(mIth{i},nt*nt,1);
    if(i~=1)
        i1 = find(x>elLim); % use for fitting only elevations above elLim (due to thread occlusions)
        i2 = find(y>elLim);
        i3 = intersect(i1,i2);
        x2 = x(i3);
        y2 = y(i3);
        z  = z(i3);
    else
        x2 = x;
        y2 = y;
    end;
    [f{i},~] = fit([x2,y2],z,'a+b/(cosd(x)*cosd(y))');
    fprintf('Thread %d params: %f %f \n',i,f{i}.a,f{i}.b);
    P.th{i}.int = [f{i}.a,f{i}.b];
end;

% fitting Phong model to specular and bcg data -----------------------
[t_i,t_v] = meshgrid(0:thetaStep:(nt-1)*thetaStep);
ti = reshape(t_i,nt*nt,1);
phi = 0*ones(nt*nt,1); 
tv = reshape(t_v,nt*nt,1);
pv = 180*ones(nt*nt,1); 
ti = [ti;reshape(t_i,nt*nt,1)];
phi = [phi;90*ones(nt*nt,1)]; 
tv = [tv;reshape(t_v,nt*nt,1)];
pv = [pv;180*ones(nt*nt,1)]; 
% compute in vector
in(:,3) = cosd(ti);
xy = sind(ti);
in(:,1) = xy.*cosd(phi);
in(:,2) = xy.*sind(phi);
in = in./repmat(sqrt(sum(in.^2,2)),1,3);
% compute out vector
out(:,3) = cosd(tv);
xy = sind(tv);
out(:,1) = xy.*cosd(pv);
out(:,2) = xy.*sind(pv);
out = out./repmat(sqrt(sum(in.^2,2)),1,3);
% compute halfway vector
half = (in+out)/2.0;
half = half./repmat(sqrt(sum(half.^2,2)),1,3);
X = half(:,3);
Y = [reshape(mIspe,nt*nt,1);reshape(mIbcg,nt*nt,1)];

beta0 = [0.2 1.0 1.0]';
options = statset('MaxIter',100,'FunValCheck','off','Robust','off','WgtFun','cauchy');
P.spe.phong = nlinfit(X,Y,@phong,beta0, options);
fprintf('Phong model: ks %f, alpha %f\n',P.spe.phong(1),P.spe.phong(2));

% plotting estimated elevation-dependendt intesitions of individual threads
Ires = zeros(nt,nt,3);
if(plot1==1)
    figure(1);
    for i=1:naxes
        subplot(2,naxes+1,i)
        imshow(1.5*uint8(scale*Ith{i}))
        title(sprintf('thread%d orig.',i));
        
        subplot(2,naxes+1,naxes+1+i)
        for j=1:3
            Ires(:,:,j) = P.th{i}.col(j)*reshape(f{i}(x,y),nt,nt);
        end;
        imshow(uint8(Ires))
        title(sprintf('thread%d restored',i));
    end;
    
    subplot(2,naxes+1,naxes+1)
    imshow(1.5*uint8(scale*Ibcg))
    title('background original');
    
    subplot(2,naxes+1,2*(naxes+1))
    C = [];
    for i=1:naxes
        C = [C;P.th{i}.col];
    end;
    imshow(uint8(permute([C;bcgCol],[1,3,2])));
    title('th 1 - th n,bcg');
end;

% =======================================================================
% modelling anisotropy ==================================================
% =======================================================================

phi_a = cell(naxes);
for j=1:naxes
    i = thOrder(j);
    phi_a{i} = (phiStep/2)*(P.th{i}.apos-1);
phi_a{i} = 360-phi_a{i};
end;

if(phi_a{1}==phi_a{2}) % if only single aHL was found
    phi_a{1} = phi_a{2}+90;
    if(phi_a{1}>360) 
        phi_a{1} = phi_a{1}-360;
    end;
end;

amap = cell(naxes);
for j=1:naxes
    i = thOrder(j);
    amap{i} = ones(np*nt,np*nt)-abs(generateAnisotropyMap(phiStep, thetaStep, phi_a{i}, AHLshift{i}));
    fprintf('Thread 1 anisotropy angle: %4.1f\n',phi_a{i});
end;

thetas = [0*ones(1,np) 15*ones(1,np) 30*ones(1,np) 45*ones(1,np) 60*ones(1,np) 75*ones(1,np)];
ti = repmat(thetas',1,np*nt);
tv = repmat(thetas ,np*nt,1);

T = cell(naxes);
for j=1:naxes
    i = thOrder(j);
    T{i} = amap{i}.*(P.th{i}.int(1) + P.th{i}.int(2)./(cosd(ti).*cosd(tv)));
end;

% carving highlights for fitting support ==================================
mT = cell(naxes);
for j=1:naxes
    i = thOrder(j);
    mT{i} = amap{i};
    mT{i}(mT{i}>w(i)) = 1;
    mT{i}(mT{i}<=w(i)) = 0;
end;

% carving highlights for rendering support ================================
iT = cell(naxes);
for j=1:naxes
    i = thOrder(j);
    iT{i} = find(mT{i}==1);
end;

% remove waves intersections before fitting ===============================
for j=1:naxes
    for i=1:naxes
        if(i~=j)
            iT{j} = setdiff(iT{j},iT{i});
        end;
    end;
end;
    
a = cell(naxes);
for j=1:naxes
    i = thOrder(j);
    a{i} = T{i}(iT{i});
end;

m = zeros(3,naxes);
phongPar = zeros(naxes,3);

beta0 = [1.0 1.0 0.2]';
options = statset('MaxIter',100,'FunValCheck','off','Robust','off','WgtFun','cauchy');

for jj=1:naxes
    i = thOrder(jj);
    phongPar(i,1:3) = nlinfit(a{i},mBRDF(iT{i}),@phongDiff,beta0, options);
    intensity = (phongPar(i,3) + phongPar(i,1).*T{i}(iT{i}).^phongPar(i,2));
    for j=1:3
        img = BRDF(:,:,j);
        b = img(iT{i});
        m(j,i) = intensity\b;
    end;
end;

% estimating background color =================================================
r = mean(P.th{thOrder(1)}.col./(P.bcg.col)); % TBD
fprintf('Ratio colors th2/bcg: %f\n',r);

toc;

% saving to parametric structure ---------------------------------
PAR.nt = naxes;
PAR.to = thOrder;
PAR.t = cell(naxes);
PAR.b = blend;
for j=1:naxes
    i = thOrder(j);
    PAR.t{i}.phi_a = phi_a{i};
    PAR.t{i}.ab = P.th{i}.int;
    PAR.t{i}.blinn = phongPar(i,:);
    PAR.t{i}.m = m(:,i)';
    PAR.t{i}.w = w(i);    
end;
PAR.r = r(1);
PAR.spe = P.spe.phong;
PAR.k = k;


fprintf('\n%s Model Parameters\n===============================\n',matName);
fid = fopen(sprintf('data/%s.txt',matName),'w');
fprintf('blend: %f\n',blend);
fprintf(fid,'%f\n',blend);
for j=1:naxes
    i = thOrder(j);
    fprintf(sprintf('THREAD %d -----------------------\n',j));
    fprintf('adir: %f\n',phi_a{i});
    fprintf('a: %f, b: %f\n', P.th{i}.int(1),P.th{i}.int(2));
    fprintf('Phong intensity: kd: %f, ks: %f, alpha: %f\n', phongPar(i,3),phongPar(i,1),phongPar(i,2));
    fprintf('R %f, G %f, B %f\n', m(1,i),m(2,i),m(3,i));
    fprintf(sprintf('w%d: %f\n',j, w(i)));
    
    fprintf(fid,'%f\n',phi_a{i});
    fprintf(fid,'%f %f\n', P.th{i}.int(1),P.th{i}.int(2));
    fprintf(fid,'%f %f %f\n', phongPar(i,3),phongPar(i,1),phongPar(i,2));
    fprintf(fid,'%f %f %f\n', m(1,i),m(2,i),m(3,i));
    fprintf(fid,'%f\n', w(i));
end;
fprintf('thread2 vs. bcg ratio: %f\n', r(1));
fprintf('Specular phong: ks: %f, alpha: %f\n', P.spe.phong(1),P.spe.phong(2));
fprintf('Specular multiplier: %f\n\n',k);
fprintf(fid,'%f\n', r(1));
fprintf(fid,'%f %f\n', P.spe.phong(1),P.spe.phong(2));
fprintf(fid,'%f\n',k);
fprintf('\n');

% restoring BRDF from model params ----------------------------------------
BRDF_r = restoreStencilBRDF(PAR,0.1);

if(plot1)
    figure(2);
    subplot(121);
    imshow(BRDF);
    title('original');
    subplot(122);
    imshow(BRDF_r);
    title('model');
end;

% computing PSNR ----------------------------------------------------------
mi = min(min(min(BRDF)));
ma = max(max(max(BRDF)));
fprintf('BRDF dyn. range: mi %f, ma %f, DR %f\n',mi,ma,ma-mi);
err = rmse(255*BRDF,255*BRDF_r, 0, ma-mi);
fprintf('RE: %f, MSE: %f, PSNR: %f, PSNR2: %f\n\n',err(1),err(3),err(4),err(5));

fprintf(fid,'RE: %f, MSE: %f, PSNR: %f, PSNR2: %f\n\n',err(1),err(3),err(4),err(5));
fclose(fid);

% saving restored BRDF ----------------------------------------------------
imwrite(BRDF_r,sprintf('data/brdf_%s_ahl.png',matName));
fid = fopen(sprintf('data/brdf_%s_ahl.bin',matName), 'w');
fwrite(fid, BRDF_r,'float');
fclose(fid);
