function BRDF_r = restoreStencilBRDF(P,smooth)

% J. Filip, UTIA AV CR v.v.i, filipj@utia.cas.cz
% v1.0, 20/6/2015
%  restoring BRDF stencils using model parameters
% 
% INPUTS: P - parameters
%         smooth - smoothing: 0 - none, 1- max (default 0.1)
% OUTPUT: BRDF_f - estimated BRDF
%
% Licence: Creative Commons Attribution-NonCommercial 4.0 International (CC BY-NC 4.0) 
%          https://creativecommons.org/licenses/by-nc/4.0/
%

% generate directions -----------------------------
phiStep = 7.5;
thetaStep = 15;
np = 360/phiStep;
nt = 90/thetaStep;

[phi_v phi_i] = meshgrid(0:np-1,0:np-1);
phi_i = phi_i*phiStep;
phi_v = phi_v*phiStep;

V = [];
for ti=0:nt-1
    for tv=0:nt-1
       theta_i = ones(np*np,1)*ti*thetaStep; 
       theta_v = ones(np*np,1)*tv*thetaStep; 
       phi_iR = reshape(phi_i,np*np,1);
       phi_vR = reshape(phi_v,np*np,1);
       V = [V;[theta_i theta_v phi_iR phi_vR]];
    end;
end;
V = V*(pi/180);
H = getHalfvector(V);

% restore threads ------------------------------------
dw = smooth;
F  = cell(P.nt);
el = cell(P.nt);
E  = cell(P.nt);
bm = cell(P.nt);
M  = cell(P.nt);
T  = cell(P.nt);

for k=1:P.nt
        
    i = P.to(k); % thread order
    F{i} = ones(size(V,1),1)-abs(H(:,1)*sind(P.t{i}.phi_a) + H(:,2)*cosd(P.t{i}.phi_a));
    el{i} = P.t{i}.ab(1) + P.t{i}.ab(2)./(cos(V(:,1)).*cos(V(:,2)));
    E{i} = F{i}.*el{i};
    bm{i} = P.t{i}.blinn(1).*E{i}.^P.t{i}.blinn(2) + P.t{i}.blinn(3);
    
    for j=1:3
        T{i}(:,j) = P.t{i}.m(j)*bm{i};
    end;
    if(smooth==0) % sharp threads translations
        M{i} = F{i};
        M{i}(M{i}<P.t{i}.w) = 0;
        M{i}(M{i}>P.t{i}.w) = 1;
    else % threads smoothing
        M{i} = smoothstep(P.t{i}.w-dw, P.t{i}.w+dw, F{i});
    end;
end;

% show individual threads contributions before combination ------------
figure(3)

for k=1:P.nt
    i = P.to(k); % thread order
    subplot(1,P.nt,i);
    aux = zeros(nt*np,nt*np,3);
    for ti=0:nt-1
        for tv=0:nt-1
            for j=1:3
                C = T{i}((ti*nt+tv)*np*np+1:(ti*nt+tv+1)*np*np,j);
                aux(ti*np+1:(ti+1)*np,tv*np+1:(tv+1)*np,j) = reshape(C',np,np);
            end;
        end;
    end;
    imshow(aux);
    title(sprintf('Thread %d',i));
end;

% combine BACKGROUND + THREAD1 + THREAD2 =================
bcg = zeros(1,3);

if(P.nt==2) % combine 2 threads
    for j=1:3
        bcg(1,j) = P.t{P.to(2)}.m(j)/P.r;
    end;
    B = repmat(bcg,size(V,1),1);
    
    for i=1:size(V,1)
        v1 = (1-M{P.to(2)}(i));
        v2 = M{P.to(2)}(i);
        w1 = (1-M{P.to(1)}(i));
        w2 = M{P.to(1)}(i);
        B(i,:) = w1 * ( v1*B(i,:) + v2*T{P.to(2)}(i,:)) +...
            w2 * ( (1-P.b(1))* T{P.to(1)}(i,:) + P.b(1)*T{P.to(2)}(i,:));
    end;
    
elseif(P.nt==3) % combine 3 threads
    to = [2 1 3]; % manual thread order: red blue green
    tb = 2; % background
    
    for j=1:3
        bcg(1,j) = P.t{to(tb)}.m(j)/P.r;
        bcg(1,j) = bcg(1,j)/1.7;
    end;
    B = repmat(bcg,size(V,1),1);
    
    for i=1:size(V,1)
        if(M{to(2)}(i)>0.0)
            B(i,:) = (1-M{to(2)}(i))*B(i,:) + M{to(2)}(i)*T{to(2)}(i,:);
        end;
        if(M{to(3)}(i)>0.0)
            B(i,:) = (1-M{to(3)}(i))*B(i,:) + M{to(3)}(i)*T{to(3)}(i,:);
        end;
        if(M{to(1)}(i)>0.0 && M{to(1)}(i)<1.0)
            B(i,:) = (1-M{to(1)}(i))*B(i,:) + M{to(1)}(i)*T{to(1)}(i,:);
        end;
        if(M{to(1)}(i)==1.0)
            B(i,:) = P.b(1)*B(i,:) + (1-P.b(1))*M{to(1)}(i)*T{to(1)}(i,:);
        end;
    end;
end;

% adding specular part ============================================
HzP = P.spe(1).*H(:,3).^P.spe(2);
HzPspe = P.k*HzP.*(P.t{P.to(1)}.ab(1) + P.t{P.to(1)}.ab(2)./(cos(V(:,1)).*cos(V(:,2))));
HzP = HzPspe;
HzP(HzP<1) = 1;
for j=1:3
    B(:,j) = B(:,j).*HzP;
end;

% reshaping array to image -----------------------------
BRDF_r = zeros(nt*np,nt*np,3);
for ti=0:nt-1
    for tv=0:nt-1
        for j=1:3
            C = B((ti*nt+tv)*np*np+1:(ti*nt+tv+1)*np*np,j);
            BRDF_r(ti*np+1:(ti+1)*np,tv*np+1:(tv+1)*np,j) = reshape(C',np,np);
        end;
    end;
end;

